"""
.. module:: cluster

cluster
*************

:Description: cluster

 Measures for clustering quality


:Authors: bejar
    

:Version: 

:Created on: 22/12/2014 13:21

# 22/09/2015: Fix: The labels from a clusterer from scikit-learn are not always contiguous

"""

__author__ = 'bejar'


import numpy as np
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.metrics.cluster.supervised import contingency_matrix, check_clusterings, mutual_info_score, entropy

def maplabels(labels):
    """
    Returns a dictionary mapping a set of labels to an index

    :param labels:
    :return:
    """
    poslabels = {}
    for lab, p in zip(labels,range(len(labels))):
        poslabels[lab] = p
    return poslabels

def scatter_matrices_scores(X, labels, indices=['CH']):
    """
    Computes different indices obtained from the Within and Between scatter matrices

    Includes:
        'SSW': Within scatter matrix score
        'SSB': Between scatter matrix score
        'Hartigan': Hartigan index
        'CH': Caliski-Harabasz index
        'Xu': Xu index
        'ZCF': ZhaoChuFranti index

    :param X:
    :param labels:
    :return:
    """
    llabels = np.unique(labels)
    poslabels = maplabels(llabels)

    nclust = len(llabels)
    nex = len(labels)

    # Centroid of the data

    centroid = np.zeros((1, X.shape[1]))
    centroid += np.sum(X, axis=0)
    centroid /= X.shape[0]

    # Compute SSB
    ccentroid = np.zeros((nclust, X.shape[1]))
    dist = 0.0
    for idx in llabels:
        center = np.zeros((1, X.shape[1]))
        center_mask = labels == idx
        center += np.sum(X[center_mask], axis=0)
        center /= center_mask.sum()
        ccentroid[poslabels[idx]] = center
        dvector = euclidean_distances(centroid, ccentroid[poslabels[idx]], squared=True)
        dist += dvector.sum() * center_mask.sum()
    SSB = dist / len(labels)

    # Compute SSW
    dist = 0.0
    for idx in llabels:
        center_mask = labels == idx
        dvector = euclidean_distances(X[center_mask], ccentroid[poslabels[idx]], squared=True)
        dist += dvector.sum()

    SSW = dist / len(labels)

    results = {}

    if 'CH' in indices:
        results['CH'] = (SSB/(nclust-1))/(SSW/(nex-nclust))

    if 'Hartigan' in indices:
        results['Hartigan'] = -np.log(SSW/SSB)

    if 'ZCF' in indices:
        results['ZCF'] = (SSW/SSB) * nclust

    if 'Xu' in indices:
        results['Xu'] = X.shape[1] * np.log(np.sqrt(SSW/(X.shape[1]*nex*nex)))+np.log(nclust)

    if 'SSW' in indices:
        results['SSW'] = SSW

    if 'SSB' in indices:
        results['SSB'] = SSB

    if 'Inertia' in indices:
        results['Inertia'] = SSW * len(labels)

    return results


def within_scatter_matrix_score(X, labels):
    """
    Computes the within scatter matrix score of a labeling of a clustering
    :param X:
    :param labels:
    :return:
    """

    llabels = np.unique(labels)

    dist = 0.0
    for idx in llabels:
        center = np.zeros((1, X.shape[1]))
        center_mask = labels == idx
        center += np.sum(X[center_mask], axis=0)
        center /= center_mask.sum()
        dvector = euclidean_distances(X[center_mask], center, squared=True)
        dist += dvector.sum()
    return dist / len(labels)


def between_scatter_matrix_score(X, labels):
    """
    Computes the between scatter matrix score of a labeling of a clustering
    :param X:
    :param labels:
    :return:
    """

    llabels = np.unique(labels)

    # Centroid of the data

    centroid = np.zeros((1, X.shape[1]))
    centroid += np.sum(X, axis=0)
    centroid /= X.shape[0]

    dist = 0.0
    for idx in llabels:
        center = np.zeros((1, X.shape[1]))
        center_mask = labels == idx
        center += np.sum(X[center_mask], axis=0)
        center /= center_mask.sum()
        dvector = euclidean_distances(centroid, center, squared=True)
        dist += dvector.sum() * center_mask.sum()
    return dist / len(labels)


def calinski_harabasz_score(X, labels):
    """
    Computes the Calinski&Harabasz score for a labeling of the data

    :param X:
    :param labels:
    :return:
    """
    llabels = np.unique(labels)
    poslabels = maplabels(llabels)

    # Centroid of the data

    centroid = np.zeros((1, X.shape[1]))
    centroid += np.sum(X, axis=0)
    centroid /= X.shape[0]

    # Compute SSB
    ccentroid = np.zeros((len(llabels), X.shape[1]))
    dist = 0.0
    for idx in llabels:
        center = np.zeros((1, X.shape[1]))
        center_mask = labels == idx
        center += np.sum(X[center_mask], axis=0)
        center /= center_mask.sum()
        ccentroid[poslabels[idx]] = center
        dvector = euclidean_distances(centroid, ccentroid[poslabels[idx]], squared=True)
        dist += dvector.sum() * center_mask.sum()
    SSB = dist / len(labels)

    # Compute SSW
    dist = 0.0
    for idx in llabels:
        center_mask = labels == idx
        dvector = euclidean_distances(X[center_mask], ccentroid[poslabels[idx]], squared=True)
        dist += dvector.sum()

    SSW = dist / len(labels)

    return (SSB/(len(llabels)-1))/(SSW/(len(labels)-len(llabels)))


def zhao_chu_franti_score(X, labels):
    """
    Implements the method defined in:

    Zhao, Q.; Xu, M. & Franti, P.  Sum-of-Squares Based Cluster Validity Index and Significance Analysis
    Adaptive and Natural Computing Algorithms, Springer Berlin Heidelberg, 2009, 5495, 313-322

    :param X:
    :param labels:
    :return:
    """
    llabels = np.unique(labels)
    poslabels = maplabels(llabels)

    # Centroid of the data

    centroid = np.zeros((1, X.shape[1]))
    centroid += np.sum(X, axis=0)
    centroid /= X.shape[0]

    # Compute SSB
    ccentroid = np.zeros((len(llabels), X.shape[1]))
    dist = 0.0
    for idx in llabels:
        center = np.zeros((1, X.shape[1]))
        center_mask = labels == idx
        center += np.sum(X[center_mask], axis=0)
        center /= center_mask.sum()
        ccentroid[poslabels[idx]] = center
        dvector = euclidean_distances(centroid, ccentroid[poslabels[idx]], squared=True)
        dist += dvector.sum() * center_mask.sum()
    SSB = dist / len(labels)

    # Compute SSW
    dist = 0.0
    for idx in llabels:
        center_mask = labels == idx
        dvector = euclidean_distances(X[center_mask], ccentroid[poslabels[idx]], squared=True)
        dist += dvector.sum()

    SSW = dist / len(labels)

    return (SSW/SSB) * len(llabels)


def davies_bouldin_score(X, labels):
    """
    Implements the Davies&Bouldin score for a labeling of the data

    :param X:
    :param labels:
    :return:
    """

    llabels = np.unique(labels)
    poslabels = maplabels(llabels)
    nclust = len(llabels)

    # compute the centroids
    centroids = np.zeros((nclust, X.shape[1]))
    for idx in llabels:
        center = np.zeros((1, X.shape[1]))
        center_mask = labels == idx
        center += np.sum(X[center_mask], axis=0)
        center /= center_mask.sum()
        centroids[poslabels[idx]] = center

    # Centroids distance matrix
    cdistances = euclidean_distances(centroids)

    # Examples to centroid mean distance
    mdcentroid = np.zeros(nclust)
    for idx in llabels:
        center_mask = labels == idx
        vdist = euclidean_distances(centroids[poslabels[idx]], X[center_mask])
        mdcentroid[poslabels[idx]] = vdist.sum()/center_mask.sum()

    # Compute the index
    dist = 0.0
    for idxi in llabels:
        lvals = []
        disti = mdcentroid[poslabels[idxi]]
        for idxj in llabels:
            if idxj != idxi:
                lvals.append((disti + mdcentroid[poslabels[idxj]])/cdistances[poslabels[idxi], poslabels[idxj]])
        dist += max(lvals)

    return dist/nclust


def jeffrey_divergence_score(X, labels):
    """
    Implements the score based on the Jeffrey divergence that appears in:

    Said, A.; Hadjidj, R. & Foufou, S. "Cluster validity index based on Jeffrey divergence"
    Pattern Analysis and Applications, Springer London, 2015, 1-11

    :param X:
    :param labels:
    :return:
    """

    llabels = np.unique(labels)
    poslabels = maplabels(llabels)
    nclust = len(llabels)

    # compute the centroids
    centroids = np.zeros((nclust, X.shape[1]))
    for idx in llabels:
        center = np.zeros((1, X.shape[1]))
        center_mask = labels == idx
        center += np.sum(X[center_mask], axis=0)
        center /= center_mask.sum()
        centroids[poslabels[idx]] = center

    lcovs = []
    linvcovs = []
    for idx in llabels:
        cov_mask = labels == idx
        covar = np.cov(X[cov_mask].T)
        lcovs.append(covar)
        linvcovs.append(np.linalg.inv(covar))

    traces = np.zeros((nclust, nclust))
    for idx1 in llabels:
        for idx2 in llabels:
            traces[poslabels[idx1], poslabels[idx2]] = np.trace(np.dot(linvcovs[poslabels[idx1]], lcovs[poslabels[idx2]]))
            traces[poslabels[idx1], poslabels[idx2]] += np.trace(np.dot(linvcovs[poslabels[idx2]], lcovs[poslabels[idx1]]))
            traces[poslabels[idx1], poslabels[idx2]] /= 2.0

    sumcov = np.zeros((nclust, nclust))
    for idx1 in llabels:
        for idx2 in llabels:
            v1 = centroids[poslabels[idx1]]
            v2 = centroids[poslabels[idx2]]
            vm = v1-v2
            mcv = linvcovs[poslabels[idx1]] + linvcovs[poslabels[idx2]]
            sumcov[poslabels[idx1], poslabels[idx2]] = np.dot(vm.T, np.dot(mcv, vm))
            sumcov[poslabels[idx1], poslabels[idx2]] /= 2.0

    ssep = 0.0
    for idx1 in llabels:
        minv = np.inf
        for idx2 in llabels:
            if idx1 != idx2:
                val = traces[poslabels[idx1], poslabels[idx2]] + sumcov[poslabels[idx1], poslabels[idx2]] - centroids.shape[1]
                if minv > val:
                    minv = val
        ssep += minv

    scompact = 0.0
    for idx in llabels:
        center_mask = labels == idx
        dvector = euclidean_distances(X[center_mask], centroids[poslabels[idx]], squared=True)
        scompact += dvector.max()

    return  scompact/ssep

def variation_of_information_score(labels_true, labels_pred):
    """Variation of Information (Meila, 2003)
    """
    labels_true, labels_pred = check_clusterings(labels_true, labels_pred)

    mutual = mutual_info_score(labels_true, labels_pred)
    e1 = entropy(labels_true)
    e2 = entropy(labels_pred)

    return e1 + e2 - (2* mutual)


def jaccard_score(labels_true, labels_pred):
    """
    Jaccard coeficient computed according to:

    Ceccarelli, M. & Maratea, A. A "Fuzzy Extension of Some Classical Concordance Measures and an Efficient Algorithm
    for Their Computation" Knowledge-Based Intelligent Information and Engineering Systems,
    Springer Berlin Heidelberg, 2008, 5179, 755-763

    :param labels_true:
    :param labels_pred:
    :return:
    """
    labels_true, labels_pred = check_clusterings(labels_true, labels_pred)
    n_samples = labels_true.shape[0]
    contingency = contingency_matrix(labels_true, labels_pred)

    cc = np.sum(contingency * contingency)
    N11 = (cc - n_samples)
    c1 = contingency.sum(axis=1)
    N01 = np.sum(c1 * c1) - cc
    c2 = contingency.sum(axis=0)
    N10 = np.sum(c2 * c2) - cc

    return (N11*1.0)/(N11+N01+N10)


def folkes_mallow_score(labels_true, labels_pred):
    """
    Folkes&Mallow score  computed according to:

    Ceccarelli, M. & Maratea, A. A "Fuzzy Extension of Some Classical Concordance Measures and an Efficient Algorithm
    for Their Computation" Knowledge-Based Intelligent Information and Engineering Systems,
    Springer Berlin Heidelberg, 2008, 5179, 755-763

    :param labels_true:
    :param labels_pred:
    :return:
    """
    labels_true, labels_pred = check_clusterings(labels_true, labels_pred)
    n_samples = labels_true.shape[0]
    contingency = contingency_matrix(labels_true, labels_pred)

    cc = np.sum(contingency * contingency)
    N11 = (cc - n_samples)
    c1 = contingency.sum(axis=1)
    N01 = np.sum(c1 * c1) - cc
    c2 = contingency.sum(axis=0)
    N10 = np.sum(c2 * c2) - cc

    return (N11*1.0)/np.sqrt((N11+N01)*(N11+N10))

def bhargavi_gowda_score(X, labels):
    """
    Score from:

    Bhargavi, M. & Gowda, S. D. "A novel validity index with dynamic cut-off for determining true clusters"
    Pattern Recognition , 2015, 48, 3673 - 3687

    :param X:
    :param labels:
    :return:
    """
    llabels = np.unique(labels)
    poslabels = maplabels(llabels)

    nclust = len(llabels)
    nex = len(labels)

    # Centroid of the data

    centroid = np.zeros((1, X.shape[1]))
    centroid += np.sum(X, axis=0)
    centroid /= X.shape[0]

    # Compute SSB and intracluster distance
    ccentroid = np.zeros((nclust, X.shape[1]))
    dist = 0.0
    for idx in llabels:
        center = np.zeros((1, X.shape[1]))
        center_mask = labels == idx
        center += np.sum(X[center_mask], axis=0)
        center /= center_mask.sum()
        ccentroid[poslabels[idx]] = center
        dvector = euclidean_distances(centroid, ccentroid[poslabels[idx]], squared=True)
        dist += dvector.sum() * center_mask.sum()

    SSB = dist / len(labels)

    # Compute SSW
    dist = 0.0
    Intra = 0.0
    for idx in llabels:
        center_mask = labels == idx
        dvector = euclidean_distances(X[center_mask], ccentroid[poslabels[idx]], squared=True)
        dist += dvector.sum()
        sdvector = euclidean_distances(X[center_mask], ccentroid[poslabels[idx]], squared=False)
        Intra += sdvector.sum()

    SSW = dist / len(labels)

    SST = SSB + SSW


    # Centroids distance matrix
    cdistances = euclidean_distances(ccentroid, squared=False)

    Inter = np.sum(cdistances)/(nclust**2)


    return(np.abs((SSW/SSB)*SST) - (Intra/Inter) - (nex - nclust))


"""
.. module:: cluster

cluster
*************

:Description: cluster

    

:Authors: bejar
    

:Version: 

:Created on: 22/12/2014 13:09 

"""

__author__ = 'bejar'

import numpy as np
from sklearn.base import BaseEstimator
from kemlglearn.metrics.cluster import within_scatter_matrix_score
from sklearn.cluster import KMeans

class quantization_error(BaseEstimator):
    """
    Method for determining the number of clusers for numerical data defined in:

    Kolesnikov, A.; Trichina, E. & Kauranne, T. (2015) "Estimating the number of clusters in a numerical data set via
    quantization error modeling" Pattern Recognition ,48, 941 - 952

    Implemented using a similar interface to the BaseEstimator class
    """
    def __init__(self, maxc=5, clustering='kmeans'):
        """

        :param maxc: Maximum number of clusters to test (and for computing the fractal dimension
        :param clustering: Clustering algorithm to use
        :return:
        """
        self._maxc = maxc
        self._clustering = clustering
        self._fdim = None
        self._M = None

    def fit(self, X):
        """

        :param X:
        :return:
        """

        lcl = range(1, self._maxc+1)

        logsum = np.log(lcl).sum()
        num = self._maxc * (np.log(lcl)*np.log(lcl)).sum()
        num -= logsum*logsum

        print(num)
        # compute the fractal dimension
        ldistorsion = []
        for i in range(1, self._maxc+1):
            cluster = KMeans(n_clusters=i, n_jobs=-1)
            cluster.fit(X)
            ldistorsion.append(cluster.inertia_/X.shape[0])

        print(ldistorsion)
        den = self._maxc * (np.log(lcl) * np.log(ldistorsion)).sum()
        den -= (logsum * np.log(ldistorsion)).sum()

        print(den)
        self._fdim = 2 * num / den

        print(self._fdim)
        PCF = [x * np.power(y, 2/self._fdim) for x, y in zip(ldistorsion, lcl)]

        print(PCF)

        self._M = np.argmin(PCF)
        print(self._M)






"""K-prototypes clustering"""

# Author: Nico de Vos <njdevos@gmail.com>
# License: MIT

from collections import defaultdict
import numpy as np
from .KModes import KModes


def euclidean_dissim(a, b):
    """Euclidean distance dissimilarity function"""
    return np.sum((a - b) ** 2, axis=1)


def move_point_num(point, ipoint, to_clust, from_clust,
                   cl_attr_sum, membership):
    """Move point between clusters, numerical attributes."""
    membership[to_clust, ipoint] = 1
    membership[from_clust, ipoint] = 0
    # Update sum of attributes in cluster.
    for iattr, curattr in enumerate(point):
        cl_attr_sum[to_clust][iattr] += curattr
        cl_attr_sum[from_clust][iattr] -= curattr
    return cl_attr_sum, membership


def _labels_cost(Xnum, Xcat, centroids, gamma):
    """Calculate labels and cost function given a matrix of points and
    a list of centroids for the k-prototypes algorithm.
    """

    npoints = Xnum.shape[0]
    cost = 0.
    labels = np.empty(npoints, dtype='int64')
    for ipoint in range(npoints):
        # Numerical cost = sum of Euclidean distances
        num_costs = euclidean_dissim(centroids[0], Xnum[ipoint])
        cat_costs = KModes.matching_dissim(centroids[1], Xcat[ipoint])
        # Gamma relates the categorical cost to the numerical cost.
        tot_costs = num_costs + gamma * cat_costs
        clust = np.argmin(tot_costs)
        labels[ipoint] = clust
        cost += tot_costs[clust]

    return labels, cost


def _k_prototypes_iter(Xnum, Xcat, centroids, cl_attr_sum, cl_attr_freq,
                       membership, gamma):
    """Single iteration of the k-prototypes algorithm"""
    moves = 0
    for ipoint in range(Xnum.shape[0]):
        clust = np.argmin(
            euclidean_dissim(centroids[0], Xnum[ipoint]) +
            gamma * KModes.matching_dissim(centroids[1], Xcat[ipoint]))
        if membership[clust, ipoint]:
            # Point is already in its right place.
            continue

        # Move point, and update old/new cluster frequencies and centroids.
        moves += 1
        old_clust = np.argwhere(membership[:, ipoint])[0][0]

        cl_attr_sum, membership = move_point_num(
            Xnum[ipoint], ipoint, clust, old_clust, cl_attr_sum,
            membership)
        cl_attr_freq, membership = KModes.move_point_cat(
            Xcat[ipoint], ipoint, clust, old_clust, cl_attr_freq,
            membership)

        # Update new and old centroids by choosing mean for numerical
        # and mode for categorical attributes.
        for iattr in range(len(Xnum[ipoint])):
            for curc in (clust, old_clust):
                if sum(membership[curc, :]):
                    centroids[0][curc, iattr] = \
                        cl_attr_sum[curc, iattr] / sum(membership[curc, :])
                else:
                    centroids[0][curc, iattr] = 0.
        for iattr in range(len(Xcat[ipoint])):
            for curc in (clust, old_clust):
                centroids[1][curc, iattr] = \
                    KModes.get_max_value_key(cl_attr_freq[curc][iattr])

        # In case of an empty cluster, reinitialize with a random point
        # from largest cluster.
        if sum(membership[old_clust, :]) == 0:
            from_clust = membership.sum(axis=1).argmax()
            choices = \
                [ii for ii, ch in enumerate(membership[from_clust, :]) if ch]
            rindx = np.random.choice(choices)

            cl_attr_freq, membership = move_point_num(
                Xnum[rindx], rindx, old_clust, from_clust, cl_attr_sum,
                membership)
            cl_attr_freq, membership = KModes.move_point_cat(
                Xcat[rindx], rindx, old_clust, from_clust, cl_attr_freq,
                membership)

    return centroids, moves


def k_prototypes(X, n_clusters, gamma, init, n_init, max_iter, verbose):
    """k-prototypes algorithm"""

    assert len(X) == 2, "X should be a list of Xnum and Xcat arrays"
    # List where [0] = numerical part of centroid and
    # [1] = categorical part. Same for centroids.
    Xnum, Xcat = X
    # Convert to numpy arrays, if needed.
    Xnum = np.asanyarray(Xnum)
    Xcat = np.asanyarray(Xcat)
    nnumpoints, nnumattrs = Xnum.shape
    ncatpoints, ncatattrs = Xcat.shape
    assert nnumpoints == ncatpoints,\
        "Uneven number of numerical and categorical points"
    npoints = nnumpoints
    assert n_clusters < npoints, "More clusters than data points?"

    # Estimate a good value for gamma, which determines the weighing of
    # categorical values in clusters (see Huang [1997]).
    if gamma is None:
        gamma = 0.5 * Xnum.std()

    all_centroids = []
    all_labels = []
    all_costs = []
    for init_no in range(n_init):

        # For numerical part of initialization, we don't have a guarantee
        # that there is not an empty cluster, so we need to retry until
        # there is none.
        while True:
            # _____ INIT _____
            if verbose:
                print("Init: initializing centroids")
            if init == 'Huang':
                centroids = KModes.init_huang(Xcat, n_clusters)
            elif init == 'Cao':
                centroids = KModes.init_cao(Xcat, n_clusters)
            elif init == 'random':
                seeds = np.random.choice(range(npoints), n_clusters)
                centroids = Xcat[seeds]
            elif hasattr(init, '__array__'):
                centroids = init
            else:
                raise NotImplementedError

            # Numerical is initialized by drawing from normal distribution,
            # categorical following the k-modes methods.
            meanX = np.mean(Xnum, axis=0)
            stdX = np.std(Xnum, axis=0)
            centroids = [meanX + np.random.randn(n_clusters, nnumattrs) * stdX,
                         centroids]

            if verbose:
                print("Init: initializing clusters")
            membership = np.zeros((n_clusters, npoints), dtype='int64')
            # Keep track of the sum of attribute values per cluster so that we
            # can do k-means on the numerical attributes.
            cl_attr_sum = np.zeros((n_clusters, nnumattrs), dtype='float')
            # cl_attr_freq is a list of lists with dictionaries that contain
            # the frequencies of values per cluster and attribute.
            cl_attr_freq = [[defaultdict(int) for _ in range(ncatattrs)]
                            for _ in range(n_clusters)]
            for ipoint in range(npoints):
                # Initial assignment to clusters
                clust = np.argmin(
                    euclidean_dissim(centroids[0], Xnum[ipoint]) +
                    gamma * KModes.matching_dissim(centroids[1], Xcat[ipoint]))
                membership[clust, ipoint] = 1
                # Count attribute values per cluster.
                for iattr, curattr in enumerate(Xnum[ipoint]):
                    cl_attr_sum[clust, iattr] += curattr
                for iattr, curattr in enumerate(Xcat[ipoint]):
                    cl_attr_freq[clust][iattr][curattr] += 1

            # If no empty clusters, then consider initialization finalized.
            if membership.sum(axis=1).min() > 0:
                break

        # Perform an initial centroid update.
        for ik in range(n_clusters):
            for iattr in range(nnumattrs):
                centroids[0][ik, iattr] =  \
                    cl_attr_sum[ik, iattr] / sum(membership[ik, :])
            for iattr in range(ncatattrs):
                centroids[1][ik, iattr] = \
                    KModes.get_max_value_key(cl_attr_freq[ik][iattr])

        # _____ ITERATION _____
        if verbose:
            print("Starting iterations...")
        itr = 0
        converged = False
        cost = np.Inf
        while itr <= max_iter and not converged:
            itr += 1
            centroids, moves = _k_prototypes_iter(
                Xnum, Xcat, centroids, cl_attr_sum, cl_attr_freq,
                membership, gamma)

            # All points seen in this iteration
            labels, ncost = \
                _labels_cost(Xnum, Xcat, centroids, gamma)
            converged = (moves == 0) or (ncost >= cost)
            cost = ncost
            if verbose:
                print("Run: {}, iteration: {}/{}, moves: {}, ncost: {}"
                      .format(init_no + 1, itr, max_iter, moves, ncost))

        # Store results of current run.
        all_centroids.append(centroids)
        all_labels.append(labels)
        all_costs.append(cost)

    best = np.argmin(all_costs)
    if n_init > 1 and verbose:
        print("Best run was number {}".format(best + 1))

    # Note: return gamma in case it was automatically determined.
    return all_centroids[best], all_labels[best], all_costs[best], gamma


class KPrototypes(KModes):
    """k-protoypes clustering algorithm for mixed numerical/categorical data.

    Parameters
    -----------
    n_clusters : int, optional, default: 8
        The number of clusters to form as well as the number of
        centroids to generate.

    gamma : float, default: None
        Weighing factor that determines relative importance of numerical vs.
        categorical attributes (see discussion in Huang [1997]). By default,
        automatically calculated from data.

    max_iter : int, default: 300
        Maximum number of iterations of the k-modes algorithm for a
        single run.

    n_init : int, default: 10
        Number of time the k-modes algorithm will be run with different
        centroid seeds. The final results will be the best output of
        n_init consecutive runs in terms of cost.

    init : {'Huang', 'Cao', 'random' or an ndarray}
        Method for initialization:
        'Huang': Method in Huang [1997, 1998]
        'Cao': Method in Cao et al. [2009]
        'random': choose k observations (rows) at random from data for
        the initial centroids.
        If an ndarray is passed, it should be of shape (n_clusters, n_features)
        and gives the initial centroids.

    verbose : boolean, optional
        Verbosity mode.

    Attributes
    ----------
    cluster_centroids_ : array, [n_clusters, n_features]
        Categories of cluster centroids

    labels_ :
        Labels of each point

    cost_ : float
        Clustering cost, defined as the sum distance of all points to
        their respective cluster centroids.

    Notes
    -----
    See:
    Huang, Z.: Extensions to the k-modes algorithm for clustering large
    data sets with categorical values, Data Mining and Knowledge
    Discovery 2(3), 1998.

    """

    def __init__(self, n_clusters=8, gamma=None, init='Huang', n_init=10,
                 max_iter=100, verbose=0):

        super(KPrototypes, self).__init__(n_clusters, init, n_init, max_iter,
                                          verbose)

        self.gamma = gamma

    def fit(self, X):
        """Compute k-prototypes clustering.

        Parameters
        ----------
        X : list of array-like, shape=[[n_num_samples, n_features],
                                       [n_cat_samples, n_features]]
        """

        # If self.gamma is None, gamma will be automatically determined from
        # the data. The function below returns its value.
        self.cluster_centroids_, self.labels_, self.cost_, self.gamma = \
            k_prototypes(X, self.n_clusters, self.gamma, self.init,
                         self.n_init, self.max_iter, self.verbose)
        return self

    def predict(self, X):
        """Predict the closest cluster each sample in X belongs to.

        Parameters
        ----------
        X : list of array-like, shape=[[n_num_samples, n_features],
                                       [n_cat_samples, n_features]]

        Returns
        -------
        labels : array, shape [n_samples,]
            Index of the cluster each sample belongs to.
        """
        assert hasattr(self, 'cluster_centroids_'), "Model not yet fitted."
        return _labels_cost(X[0], X[1], self.cluster_centroids_,
                            self.gamma)[0]



"""K-modes clustering"""

# Author: Nico de Vos <njdevos@gmail.com>
# License: MIT

from collections import defaultdict
import numpy as np


def get_max_value_key(dic):
    """Fast method to get key for maximum value in dict."""
    v = list(dic.values())
    k = list(dic.keys())
    return k[v.index(max(v))]


def matching_dissim(a, b):
    """Simple matching dissimilarity function"""
    return np.sum(a != b, axis=1)


def init_huang(X, n_clusters):
    """Initialize n_clusters according to method by Huang [1997]."""
    nattrs = X.shape[1]
    centroids = np.empty((n_clusters, nattrs), dtype='object')
    # determine frequencies of attributes
    for iattr in range(nattrs):
        freq = defaultdict(int)
        for curattr in X[:, iattr]:
            freq[curattr] += 1
        # Sample centroids using the probabilities of attributes.
        # (I assume that's what's meant in the Huang [1998] paper; it works,
        # at least)
        # Note: sampling using population in static list with as many choices
        # as frequency counts. Since the counts are small integers,
        # memory consumption is low.
        choices = [chc for chc, wght in freq.items() for _ in range(wght)]
        centroids[:, iattr] = np.random.choice(choices, n_clusters)
    # The previously chosen centroids could result in empty clusters,
    # so set centroid to closest point in X.
    for ik in range(n_clusters):
        ndx = np.argsort(matching_dissim(X, centroids[ik]))
        # We want the centroid to be unique.
        while np.all(X[ndx[0]] == centroids, axis=1).any():
            ndx = np.delete(ndx, 0)
        centroids[ik] = X[ndx[0]]

    return centroids


def init_cao(X, n_clusters):
    """Initialize n_clusters according to method by Cao et al. [2009].

    Note: O(N * attr * n_clusters**2), so watch out with large n_clusters
    """
    npoints, nattrs = X.shape
    centroids = np.empty((n_clusters, nattrs), dtype='object')
    # Method is base don determining density of points.
    dens = np.zeros(npoints)
    for iattr in range(nattrs):
        freq = defaultdict(int)
        for val in X[:, iattr]:
            freq[val] += 1
        for ipoint in range(npoints):
            dens[ipoint] += freq[X[ipoint, iattr]] / float(nattrs)
    dens /= npoints

    # Choose initial centroids based on distance and density.
    centroids[0] = X[np.argmax(dens)]
    if n_clusters > 1:
        # For the remaining centroids, choose maximum dens * dissim to the
        # (already assigned) centroid with the lowest dens * dissim.
        for ik in range(1, n_clusters):
            dd = np.empty((ik, npoints))
            for ikk in range(ik):
                dd[ikk] = matching_dissim(X, centroids[ikk]) * dens
            centroids[ik] = X[np.argmax(np.min(dd, axis=0))]

    return centroids


def move_point_cat(point, ipoint, to_clust, from_clust,
                   cl_attr_freq, membership):
    """Move point between clusters, categorical attributes."""
    membership[to_clust, ipoint] = 1
    membership[from_clust, ipoint] = 0
    # Update frequencies of attributes in cluster.
    for iattr, curattr in enumerate(point):
        cl_attr_freq[to_clust][iattr][curattr] += 1
        cl_attr_freq[from_clust][iattr][curattr] -= 1
    return cl_attr_freq, membership


def _labels_cost(X, centroids):
    """Calculate labels and cost function given a matrix of points and
    a list of centroids for the k-modes algorithm.
    """

    npoints = X.shape[0]
    cost = 0.
    labels = np.empty(npoints, dtype='int64')
    for ipoint, curpoint in enumerate(X):
        diss = matching_dissim(centroids, curpoint)
        clust = np.argmin(diss)
        labels[ipoint] = clust
        cost += diss[clust]

    return labels, cost


def _k_modes_iter(X, centroids, cl_attr_freq, membership):
    """Single iteration of k-modes clustering algorithm"""
    moves = 0
    for ipoint, curpoint in enumerate(X):
        clust = np.argmin(matching_dissim(centroids, curpoint))
        if membership[clust, ipoint]:
            # Point is already in its right place.
            continue

        # Move point, and update old/new cluster frequencies and centroids.
        moves += 1
        old_clust = np.argwhere(membership[:, ipoint])[0][0]

        cl_attr_freq, membership = move_point_cat(
            curpoint, ipoint, clust, old_clust, cl_attr_freq, membership)

        # Update new and old centroids by choosing mode of attribute.
        for iattr in range(len(curpoint)):
            for curc in (clust, old_clust):
                centroids[curc, iattr] = \
                    get_max_value_key(cl_attr_freq[curc][iattr])

        # In case of an empty cluster, reinitialize with a random point
        # from the largest cluster.
        if sum(membership[old_clust, :]) == 0:
            from_clust = membership.sum(axis=1).argmax()
            choices = \
                [ii for ii, ch in enumerate(membership[from_clust, :]) if ch]
            rindx = np.random.choice(choices)

            cl_attr_freq, membership = move_point_cat(
                X[rindx], rindx, old_clust, from_clust, cl_attr_freq,
                membership)

    return centroids, moves


def k_modes(X, n_clusters, init, n_init, max_iter, verbose):
    """k-modes algorithm"""

    # Convert to numpy array, if needed.
    X = np.asanyarray(X)
    npoints, nattrs = X.shape
    assert n_clusters < npoints, "More clusters than data points?"

    all_centroids = []
    all_labels = []
    all_costs = []
    for init_no in range(n_init):

        # _____ INIT _____
        if verbose:
            print("Init: initializing centroids")
        if init == 'Huang':
            centroids = init_huang(X, n_clusters)
        elif init == 'Cao':
            centroids = init_cao(X, n_clusters)
        elif init == 'random':
            seeds = np.random.choice(range(npoints), n_clusters)
            centroids = X[seeds]
        elif hasattr(init, '__array__'):
            centroids = init
        else:
            raise NotImplementedError

        if verbose:
            print("Init: initializing clusters")
        membership = np.zeros((n_clusters, npoints), dtype='int64')
        # cl_attr_freq is a list of lists with dictionaries that contain the
        # frequencies of values per cluster and attribute.
        cl_attr_freq = [[defaultdict(int) for _ in range(nattrs)]
                        for _ in range(n_clusters)]
        for ipoint, curpoint in enumerate(X):
            # Initial assignment to clusters
            clust = np.argmin(matching_dissim(centroids, curpoint))
            membership[clust, ipoint] = 1
            # Count attribute values per cluster.
            for iattr, curattr in enumerate(curpoint):
                cl_attr_freq[clust][iattr][curattr] += 1
        # Perform an initial centroid update.
        for ik in range(n_clusters):
            for iattr in range(nattrs):
                centroids[ik, iattr] = get_max_value_key(
                    cl_attr_freq[ik][iattr])

        # _____ ITERATION _____
        if verbose:
            print("Starting iterations...")
        itr = 0
        converged = False
        cost = np.Inf
        while itr <= max_iter and not converged:
            itr += 1
            centroids, moves = \
                _k_modes_iter(X, centroids, cl_attr_freq, membership)
            # All points seen in this iteration
            labels, ncost = _labels_cost(X, centroids)
            converged = (moves == 0) or (ncost >= cost)
            cost = ncost
            if verbose:
                print("Run {}, iteration: {}/{}, moves: {}, cost: {}"
                      .format(init_no + 1, itr, max_iter, moves, cost))

        # Store result of current run.
        all_centroids.append(centroids)
        all_labels.append(labels)
        all_costs.append(cost)

    best = np.argmin(all_costs)
    if n_init > 1 and verbose:
        print("Best run was number {}".format(best + 1))

    return all_centroids[best], all_labels[best], all_costs[best]


class KModes(object):

    """k-modes clustering algorithm for categorical data.

    Parameters
    -----------
    n_clusters : int, optional, default: 8
        The number of clusters to form as well as the number of
        centroids to generate.

    max_iter : int, default: 300
        Maximum number of iterations of the k-modes algorithm for a
        single run.

    n_init : int, default: 10
        Number of time the k-modes algorithm will be run with different
        centroid seeds. The final results will be the best output of
        n_init consecutive runs in terms of cost.

    init : {'Huang', 'Cao', 'random' or an ndarray}
        Method for initialization:
        'Huang': Method in Huang [1997, 1998]
        'Cao': Method in Cao et al. [2009]
        'random': choose k observations (rows) at random from data for
        the initial centroids.
        If an ndarray is passed, it should be of shape (n_clusters, n_features)
        and gives the initial centroids.

    verbose : boolean, optional
        Verbosity mode.

    Attributes
    ----------
    cluster_centroids_ : array, [n_clusters, n_features]
        Categories of cluster centroids

    labels_ :
        Labels of each point

    cost_ : float
        Clustering cost, defined as the sum distance of all points to
        their respective cluster centroids.

    Notes
    -----
    See:
    Huang, Z.: Extensions to the k-modes algorithm for clustering large
    data sets with categorical values, Data Mining and Knowledge
    Discovery 2(3), 1998.

    """

    def __init__(self, n_clusters=8, init='Cao', n_init=10, max_iter=100,
                 verbose=0):

        if hasattr(init, '__array__'):
            n_clusters = init.shape[0]
            init = np.asarray(init, dtype=np.float64)

        self.n_clusters = n_clusters
        self.init = init
        self.n_init = n_init
        self.verbose = verbose
        if (self.init == 'Cao' or hasattr(self.init, '__array__')) and \
                self.n_init > 1:
            if self.verbose:
                print("Initialization method and algorithm are deterministic. "
                      "Setting n_init to 1.")
            self.n_init = 1

        self.max_iter = max_iter

    def fit(self, X):
        """Compute k-modes clustering.

        Parameters
        ----------
        X : array-like, shape=[n_samples, n_features]
        """

        self.cluster_centroids_, self.labels_, self.cost_ = \
            k_modes(X, self.n_clusters, self.init, self.n_init,
                    self.max_iter, self.verbose)
        return self

    def fit_predict(self, X, y=None):
        """Compute cluster centroids and predict cluster index for each sample.

        Convenience method; equivalent to calling fit(X) followed by
        predict(X).
        """
        return self.fit(X).labels_

    def predict(self, X):
        """Predict the closest cluster each sample in X belongs to.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            New data to predict.

        Returns
        -------
        labels : array, shape [n_samples,]
            Index of the cluster each sample belongs to.
        """
        assert hasattr(self, 'cluster_centroids_'), "Model not yet fitted."
        return _labels_cost(X, self.cluster_centroids_)[0]



"""
.. module:: samples_generator

samples_generator
*************

:Description: samples_generator

    

:Authors: bejar
    

:Version: 

:Created on: 21/01/2015 9:02 

"""

__author__ = 'bejar'

from rpy2.robjects.packages import importr
import numpy as np
import numbers
from sklearn.utils import check_random_state, check_array


def make_blobs(n_samples=100, n_features=2, centers=3, cluster_std=1.0,
               center_box=(-10.0, 10.0), shuffle=True, random_state=None):
    """Generate isotropic Gaussian blobs for clustering.

    7/10/2015
    A fixed and more flexible version of the scikit-learn function

    Parameters
    ----------
    n_samples : int, or sequence of integers, optional (default=100)
        The total number of points equally divided among clusters.
        or a sequence of the number of examples of each cluster

    n_features : int, optional (default=2)
        The number of features for each sample.

    centers : int or array of shape [n_centers, n_features], optional
        (default=3)
        The number of centers to generate, or the fixed center locations.

    cluster_std: float or sequence of floats, optional (default=1.0)
        The standard deviation of the clusters.
        now works for the list of floats

    center_box: pair of floats (min, max), optional (default=(-10.0, 10.0))
        The bounding box for each cluster center when centers are
        generated at random.

    shuffle : boolean, optional (default=True)
        Shuffle the samples.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Returns
    -------
    X : array of shape [n_samples, n_features]
        The generated samples.

    y : array of shape [n_samples]
        The integer labels for cluster membership of each sample.

    Examples
    --------
    >>> from sklearn.datasets.samples_generator import make_blobs
    >>> X, y = make_blobs(n_samples=10, centers=3, n_features=2,
    ...                   random_state=0)
    >>> print(X.shape)
    (10, 2)
    >>> y
    array([0, 0, 1, 0, 2, 2, 2, 1, 1, 0])
    """
    generator = check_random_state(random_state)

    if isinstance(centers, numbers.Integral):
        centers = generator.uniform(center_box[0], center_box[1],
                                    size=(centers, n_features))
    else:
        centers = check_array(centers)
        n_features = centers.shape[1]

    X = []
    y = []

    n_centers = centers.shape[0]
    if not isinstance(n_samples, list):
        n_samples_per_center = [int(n_samples // n_centers)] * n_centers
        for i in range(n_samples % n_centers):
            n_samples_per_center[i] += 1
    else:
        if len(n_samples) != n_centers:
            raise NameError('List of number of examples per center doer not match number of centers')
        n_samples_per_center = n_samples
        n_samples = sum(n_samples)

    if not isinstance(cluster_std, list):
        std_list = [cluster_std] * centers.shape[0]
    else:
        if len(cluster_std) != n_centers:
            raise NameError('List of number of examples per center doer not match number of centers')
        std_list = cluster_std

    for i, (n, st) in enumerate(zip(n_samples_per_center, std_list)):
        X.append(centers[i] + generator.normal(scale=st,
                                               size=(n, n_features)))
        y += [i] * n

    X = np.concatenate(X)
    y = np.array(y)

    if shuffle:
        indices = np.arange(n_samples)
        generator.shuffle(indices)
        X = X[indices]
        y = y[indices]

    return X, y


def cluster_generator(n_clusters=3, sepval=0.5, numNonNoisy=5, numNoisy=0, numOutlier=0,
                      clustszind=2, clustSizeEq=100, rangeN=[100, 150], rotateind=True):
    """
    Generates clusters using the R package CusterGeneration
    See the documentation of that package for the meaning of the parameters

    You must have an R installation with the clusterGeneration package

    :param n_clusters:
    :param sepval:
    :return:
    """
    clusterG = importr('clusterGeneration')

    params = {'numClust': n_clusters,
          'sepVal': sepval,
          'numNonNoisy': numNonNoisy,
          'numNoisy': numNoisy,
          'numOutlier': numOutlier,
          'numReplicate': 1,
          'clustszind': clustszind,
          'clustSizeEq': clustSizeEq,
          'rangeN': rangeN,
          'rotateind': rotateind,
          'outputDatFlag': False,
          'outputLogFlag': False,
          'outputEmpirical': False,
          'outputInfo': False
         }

    x = clusterG.genRandomClust(**params)
    # nm = np.array(x[2][0].colnames)
    # nm = np.concatenate((nm, ['class']))
    m = np.matrix(x[2][0])
    v = np.array(x[3][0])
    v.resize((len(x[3][0])))
    #m = np.concatenate((m, v), axis=1)
    return m, v



"""
.. module:: LaplacianScore

LaplacianScore
*************

:Description: LaplacianScore

    Class that computes the laplacian score for a dataset

:Authors: bejar
    

:Version: 

:Created on: 25/11/2014 9:32 

"""

__author__ = 'bejar'

import numpy as np
from sklearn.neighbors import kneighbors_graph, NearestNeighbors
from operator import itemgetter

class LaplacianScore():
    """
    Laplacian Score algorithm

    Parameters

        - Number of neighbors to compute the similarity matrix
        - Bandwidth for the gaussian similarity kernel
    """

    scores_ = None


    def __init__(self, n_neighbors=5, bandwidth=0.01, k=None):
        """
        Initial values of the parameters
        @param n_neighbors: Number of neighbors for the spectral matrix
        @param bandwidth: Bandwidth for the gaussian kernel
        @param k: number of features to select
        @return:
        """
        self._n_neighbors = n_neighbors
        self._bandwidth = bandwidth
        self._k = k

    def fit(self, X):
        """
        Computes the laplacian scores for the dataset

        X is a [n_examples, n_attributes] numpy array
        @return:
        """

        self._fit_process(X)

        return self

    def _best_k_scores(self, k=5):
        """
        returns the indices of the best k attributes according to the score

        :param k:
        :return:
        """
        if self.scores_ is None:
            raise Exception('Laplacian Score: Not fitted')
        else:
            l = list(enumerate(self.scores_))
            l = sorted(l, key=itemgetter(1), reverse=True)
            return [l[i][0] for i in range(k)]

    def fit_transform(self, X):
        """
        Selects the features and returns the dataset with only the k best ones
        :return:
        """

        self._fit_process(X)
        l = list(enumerate(self.scores_))
        l = sorted(l, key=itemgetter(1), reverse=True)
        lsel = [l[i][0] for i in range(self._k)]
        return X[:, lsel]

    # Todo: implementation only with sparse matrices
    def _fit_process(self, X):
        """
        Computes the Laplacian score for the attributes


        @param X:
        @return:
        """

        self.scores_ = np.zeros(X.shape[1])

        # Similarity matrix
        S = kneighbors_graph(X, n_neighbors=self._n_neighbors, mode='distance')
        S = S.toarray()
        S *= S
        S /= self._bandwidth
        S = -S

        ones = np.ones(X.shape[0])

        D = np.diag(np.dot(S, ones))

        L = D - S

        qt = D.sum()
        for at in range(X.shape[1]):
            Fr = X[:, at]
            Fr_hat = Fr - np.dot(np.dot(Fr, D) / qt, ones)

            score1 = np.dot(np.dot(Fr_hat, L), Fr_hat)
            score2 = np.dot(np.dot(Fr_hat, D), Fr_hat)
            self.scores_[at] = score1 / score2







"""
.. module:: SimpleConsensusClustering

SimpleConsensusClustering
*************

:Description: SimpleConsensusClustering

    

:Authors: bejar
    

:Version: 

:Created on: 22/01/2015 10:46 

"""

__author__ = 'bejar'


import numpy as np
from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.cluster import KMeans, SpectralClustering


class SimpleConsensusClustering(BaseEstimator, ClusterMixin, TransformerMixin):
    """Simple Consensus Clustering Algorithm

    Pararemeters:

    n_clusters: int
        Number of clusters of the base clusterers and the consensus cluster
    base: string
        base clusterer ['kmeans']
    n_components: int
        number of components of the consensus
    consensus: string
        consensus method ['coincidence']

    """
    def __init__(self, n_clusters, base='kmeans', n_components=10,
                 consensus='coincidence', consensus2='kmeans'):
        self.n_clusters = n_clusters
        self.cluster_centers_ = None
        self.labels_ = None
        self.cluster_sizes_ = None
        self.base = base
        self.n_components = n_components
        self.consensus = consensus
        self.consensus2 = consensus2


    def fit(self, X):
        """
        Clusters the examples
        :param X:
        :return:
        """

        if self.consensus == 'coincidence':
            self.cluster_centers_, self.labels_ = self._fit_process_coincidence(X)


    def _fit_process_coincidence(self, X):
        """
        Obtains n_components kmeans clustering, compute the coincidence matrix and applies kmeans to that coincidence
        matrix

        :param X:
        :return:
        """
        baseclust = []
        if self.base == 'kmeans':
            km = KMeans(n_clusters=self.n_clusters, n_jobs=-1)
        elif self.base == 'spectral':
            km = SpectralClustering(n_clusters=self.n_clusters, assign_labels='discretize',
                                    affinity='nearest_neighbors', n_neighbors=30)

        for i in range(self.n_components):
            km.fit(X)
            baseclust.append(km.labels_)

        coin_matrix = np.zeros((X.shape[0], X.shape[0]))

        for l in baseclust:
            for i in range(X.shape[0]):
                for j in range(i+1, X.shape[0]):
                    if i != j:
                        if l[i] == l[j]:
                            coin_matrix[i, j] += 1
                            coin_matrix[j, i] += 1

        coin_matrix /= self.n_components
        if self.consensus2 == 'kmeans':
            kmc = KMeans(n_clusters=self.n_clusters, n_jobs=-1)
            kmc.fit(coin_matrix)
            return kmc.cluster_centers_, kmc.labels_
        elif self.consensus2 == 'spectral':
            kmc = SpectralClustering(n_clusters=self.n_clusters, assign_labels='discretize',
                                    affinity='nearest_neighbors', n_neighbors=40)
            kmc.fit(coin_matrix)

            return None, kmc.labels_





"""
.. module:: MeanPartition

MeanPartition
*************

:Description: MeanPartition

    

:Authors: bejar
    

:Version: 

:Created on: 10/02/2015 9:40 

"""

__author__ = 'bejar'

import numpy as np
from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.cluster import KMeans, SpectralClustering
from sklearn.metrics import adjusted_mutual_info_score, adjusted_rand_score, v_measure_score
from sklearn.manifold import MDS, TSNE, SpectralEmbedding

class MeanPartitionClustering(BaseEstimator, ClusterMixin, TransformerMixin):
    """Consensus Clustering Algorithm based on the estimation of the mean partition
    """
    def __init__(self, n_clusters, base='kmeans', n_components=10, cdistance='ANMI', trans='spectral', n_neighbors=5):
        self.n_clusters = n_clusters
        self.cluster_centers_ = None
        self.labels_ = None
        self.cluster_sizes_ = None
        self.base = base
        self.n_components = n_components
        self.trans = trans
        self.cdistance = cdistance
        self.n_neighbors = n_neighbors

    def fit(self, X, y):
        """
        Clusters the examples
        :param X:
        :return:
        """
        baseclust = [y]
        if self.base == 'kmeans':
            km = KMeans(n_clusters=self.n_clusters, n_jobs=-1)
        l = ['r']

        for i in range(self.n_components):
            km.fit(X)
            baseclust.append(km.labels_)
            l.append('b')

        mdist = np.zeros((self.n_components, self.n_components))
        if self.cdistance == 'ANMI':
            dfun = adjusted_mutual_info_score
        elif self.cdistance == 'ARAND':
            dfun = adjusted_rand_score
        elif self.cdistance == 'vmeasure':
            dfun = v_measure_score

        for i in range(self.n_components-1):
            for j in range(i+1, self.n_components):
                mdist[i, j] = dfun(baseclust[i], baseclust[j])
                mdist[j, i] = mdist[i, j]


        if self.trans == 'MDS':
            embed = MDS(dissimilarity='precomputed')
        elif self.trans == 'TNE':
            embed = TSNE(metric='precomputed')
        elif self.trans == 'spectral':
            embed = SpectralEmbedding(affinity='precomputed', n_neighbors=self.n_neighbors)

        X = embed.fit_transform(mdist)

        return X, l



"""
.. module:: BorderDetection

BorderDetection
*************

:Description: BorderDetection

    Algorithms for cluster border detection

:Authors: bejar
    

:Version: 

:Created on: 16/06/2016 9:25 

"""

__author__ = 'bejar'

from sklearn.neighbors import NearestNeighbors
import numpy as np

def QiuCaoBorder(X, n_neighbors, boundary_ratio, filter_ratio):
    """
    Compute the border points of the clusters from the dataset as described in:

    Qiu, B. & Cao, X. Clustering boundary detection for high dimensional space based on space
    inversion and Hopkins statistics Knowledge-Based Systems , 2016, 98, 216 - 225

    :return:
    """
    nn = NearestNeighbors(n_neighbors=n_neighbors+1)
    nn.fit(X)

    Xsym = np.zeros(X.shape[0])
    Filter = np.zeros(X.shape[0])

    for i in range(X.shape[0]):
        dnn, lnn = nn.kneighbors(X[i,:].reshape(1, -1))
        PLocations = np.zeros((n_neighbors, X.shape[1]))
        sdist = 0
        for j in range(n_neighbors):
            PLocations[j] = X[lnn[0][j+1]] - X[lnn[0][0]]
            sdist += np.exp(dnn[0][j+1]/n_neighbors)

        vsym = PLocations.sum(axis=1)
        Xsym[i] = np.abs(vsym).sum()
        Filter[i] = sdist

    eps1 = sorted(Xsym, reverse=True)[int(boundary_ratio*X.shape[0])]
    eps2 = sorted(Filter, reverse=True)[int(filter_ratio*X.shape[0])]

    border = []
    for i in range(X.shape[0]):
        if Xsym[i] > eps1 and Filter[i] < eps2:
            border.append(i)

    return border

if __name__ == '__main__':
    from sklearn.datasets import make_blobs, load_iris, make_circles
    import pylab as pl
    import matplotlib.pyplot as plt

    #X = load_iris()['data']

    X, y_data = make_blobs(n_samples=10000, n_features=5, centers=5, cluster_std=0.5, random_state=2)

    bd = QiuCaoBorder(X, 5, 0.2, 0.01)
    print bd
    labels = np.zeros(X.shape[0])+1
    for i in bd:
        labels[i] = 5
    fig = plt.figure()
    plt.scatter(X[:,0],X[:,1], s=labels*10)
    plt.show()












"""
.. module:: __init__.py

__init__.py
*************

:Description: __init__.py

    

:Authors: bejar
    

:Version: 

:Created on: 22/01/2015 10:46 

"""

__author__ = 'bejar'

from .SimpleConsensusClustering import SimpleConsensusClustering
from .MeanPartition import MeanPartitionClustering

__all__ = ['SimpleConsensusClustering',
           'MeanPartitionClustering']



"""
.. module:: test

test
*************

:Description: test

    

:Authors: bejar
    

:Version: 

:Created on: 13/03/2015 16:34 

"""

__author__ = 'bejar'

from kemlglearn.preprocessing import Discretizer

from sklearn.datasets import make_blobs, load_iris, make_circles

X = load_iris()['data']

print(X.shape)

disc = Discretizer(bins=3, method='frequency')

disc.fit(X)

print(disc.intervals)

disc = Discretizer(bins=3, method='equal')

disc.fit(X)

print(disc.intervals)


# Y = disc.transform(X, copy=True)
#
#
# for i in range(X.shape[0]):
#     print X[i,:], Y[i,:]



"""
.. module:: nestedpartition

nestedpartition
*************

:Description: nestedpartition

    

:Authors: bejar
    

:Version: 

:Created on: 09/07/2015 11:18 

"""

__author__ = 'bejar'


import numpy as np
import pprint

def nested_item(depth, value):
    if depth <= 1:
        print(value)
        return [value]
    else:
        return [nested_item(depth - 1, value)]

def nested_list(n):
    """Generate a nested list where the i'th item is at depth i."""
    lis = []
    for i in range(n):
        if i == 0:
            lis.append(i)
        else:
            lis.append(nested_item(i, i))
    return lis

def flatten(lis):
    """Given a list, possibly nested to any level, return it flattened."""
    new_lis = []
    for item in lis:
        if type(item) == type([]):
            new_lis.extend(flatten(item))
        else:
            new_lis.append(item)
    return new_lis



def nested_partitions_distance(p1, p2):
    """
    Distance among nested partitions represented as nested lists

    :param p1:
    :param p2:
    :return:

    """

    if len(p1) >1 and len(p2) > 1:
        set1a = set(flatten(p1[0]))
        set1b = set(flatten(p1[1]))
        set2a = set(flatten(p2[0]))
        set2b = set(flatten(p2[1]))
        if len(set1a.intersection(set2a)) > len(set1a.intersection(set2b)):
            return len(set1a.intersection(set2a)) +\
                   len(set1b.intersection(set2b)) +\
                   nested_partitions_distance(p1[0], p2[0]) +\
                   nested_partitions_distance(p1[1], p2[1])
        else:
            return len(set1a.intersection(set2b)) +\
                   len(set1b.intersection(set2a)) +\
                   nested_partitions_distance(p1[0], p2[1]) +\
                   nested_partitions_distance(p1[1], p2[0])
    else:
        if len(p1) == 1 and len(p2) > 1:
            if p1[0] in set(flatten(p2[0])) or p1[0] in set(flatten(p2[1])):
                return 1
            else:
                return 0
        elif len(p2) == 1 and len(p1) > 1:
            if p2[0] in set(flatten(p1[0])) or p2[0] in set(flatten(p1[1])):
                return 1
            else:
                return 0
        else:
            if p1[0] == p2[0]:
                return 1
            else:
                return 0


def nested_partitions_distance2(p1, p2):
    """
    Distance among nested partitions represented as nested lists

    :param p1:
    :param p2:
    :return:

    """
    if len(p1) >1 and len(p2) > 1:
        set1a = set(flatten(p1[0]))
        set1b = set(flatten(p1[1]))
        set2a = set(flatten(p2[0]))
        set2b = set(flatten(p2[1]))
        if len(set1a.intersection(set2a)) > len(set1a.intersection(set2b)):
            return nested_partitions_distance(p1[0], p2[0]) +\
                   nested_partitions_distance(p1[1], p2[1])
        else:
            return nested_partitions_distance(p1[0], p2[1]) +\
                   nested_partitions_distance(p1[1], p2[0])
    else:
        if len(p1) == 1 and len(p2) > 1:
            if p1[0] in set(flatten(p2[0])) or p1[0] in set(flatten(p2[1])):
                return 1
            else:
                return 0
        elif len(p2) == 1 and len(p1) > 1:
            if p2[0] in set(flatten(p1[0])) or p2[0] in set(flatten(p1[1])):
                return 1
            else:
                return 0
        else:
            if p1[0] == p2[0]:
                return 1
            else:
                return 0


def generate_partitions(data):
    """
    Generates a random nested partition for an array of integers

    :param data:
    :return:
    """
    if len(data) == 1:
        return data
    else:

        mask1 = np.random.choice(len(data), np.floor(len(data)/2), replace=False)
        par1 = [data[i] for i in range(len(data)) if i in mask1]
        par2 = [data[i] for i in range(len(data)) if i not in mask1]
        return [generate_partitions(par1), generate_partitions(par2)]

def print_nested(l,p):
    if len(l) == 1:
        print(' '*p, l)
    else:
        for v in l:
            print_nested(v, p+1)

# ---------------------------------------------------------------------------------------------------------------
if __name__ == '__main__':
    pp = pprint.PrettyPrinter(indent=4)
    a = np.array(range(100))

    b = generate_partitions(a)
    c = generate_partitions(a)
    # print_nested(b,0)
    # print('---')
    # print_nested(c,0)
    print(nested_partitions_distance2(b, c))








"""
.. module:: Xu

Xu
*************

:Description: Xu

    

:Authors: bejar
    

:Version: 

:Created on: 22/12/2014 15:26 

"""

__author__ = 'bejar'


import numpy as np
from sklearn.base import BaseEstimator
from kemlglearn.metrics.cluster import within_scatter_matrix_score
from sklearn.cluster import KMeans

class Xu(BaseEstimator):
    """
    Method for determining the number of clusters for numerical data defined in:

    XU, L. (1997) Bayesian Ying-Yang Machine, clustering and Number of Clusters.
    Pattern Recognition Letters 18, 1167-1178

    Implemented using a similar interface to the BaseEstimator class
    """
    def __init__(self, maxc=5, clustering='kmeans'):
        """

        :param maxc: Maximum number of clusters to test (and for computing the fractal dimension
        :param clustering: Clustering algorithm to use
        :return:
        """
        self._maxc = maxc
        self._clustering = clustering
        self._M = None

    def fit(self, X):
        """

        :param X:
        :return:
        """

        lcl = range(1, self._maxc+1)


        # compute the fractal dimension
        ldistorsion = []
        for i in range(1, self._maxc+1):
            cluster = KMeans(n_clusters=i, n_jobs=-1)
            cluster.fit(X)
            ldistorsion.append(within_scatter_matrix_score(X, cluster.labels_))

        print(X.shape[1])
        print(ldistorsion)

        PCF = []
        for x,y in zip(ldistorsion, lcl):
            print(x,y, np.power(y, 2.0/X.shape[1]))
            PCF.append(x * np.power(y, 2.0/X.shape[1]))

        print(PCF)

        self._M = np.argmin(PCF)
        print(self._M)



"""
.. module:: KernelKMeans

KernekKMeans
*************

:Description: KernelKMeans


"""

# Author: Mathieu Blondel <mathieu@mblondel.org>
# License: BSD 3 clause

import numpy as np

from sklearn.base import BaseEstimator, ClusterMixin
from sklearn.metrics.pairwise import pairwise_kernels
from sklearn.utils import check_random_state


class KernelKMeans(BaseEstimator, ClusterMixin):
    """
    Kernel K-means

    Reference
    ---------
    Kernel k-means, Spectral Clustering and Normalized Cuts.
    Inderjit S. Dhillon, Yuqiang Guan, Brian Kulis.
    KDD 2004.
    """

    def __init__(self, n_clusters=3, max_iter=50, tol=1e-3, random_state=None,
                 kernel="linear", gamma=None, degree=3, coef0=1,
                 kernel_params=None, verbose=0):
        self.n_clusters = n_clusters
        self.max_iter = max_iter
        self.tol = tol
        self.random_state = random_state
        self.kernel = kernel
        self.gamma = gamma
        self.degree = degree
        self.coef0 = coef0
        self.kernel_params = kernel_params
        self.verbose = verbose

    @property
    def _pairwise(self):
        return self.kernel == "precomputed"

    def _get_kernel(self, X, Y=None):
        if callable(self.kernel):
            params = self.kernel_params or {}
        else:
            params = {"gamma": self.gamma,
                      "degree": self.degree,
                      "coef0": self.coef0}
        return pairwise_kernels(X, Y, metric=self.kernel,
                                filter_params=True, **params)

    def fit(self, X, y=None, sample_weight=None):
        n_samples = X.shape[0]

        K = self._get_kernel(X)

        sw = sample_weight if sample_weight else np.ones(n_samples)
        self.sample_weight_ = sw

        rs = check_random_state(self.random_state)
        self.labels_ = rs.randint(self.n_clusters, size=n_samples)

        dist = np.zeros((n_samples, self.n_clusters))
        self.within_distances_ = np.zeros(self.n_clusters)

        for it in xrange(self.max_iter):
            dist.fill(0)
            self._compute_dist(K, dist, self.within_distances_,
                               update_within=True)
            labels_old = self.labels_
            self.labels_ = dist.argmin(axis=1)

            # Compute the number of samples whose cluster did not change
            # since last iteration.
            n_same = np.sum((self.labels_ - labels_old) == 0)
            if 1 - float(n_same) / n_samples < self.tol:
                if self.verbose:
                    print("Converged at iteration", it + 1)
                break

        self.X_fit_ = X

        return self

    def _compute_dist(self, K, dist, within_distances, update_within):
        """Compute a n_samples x n_clusters distance matrix using the
        kernel trick."""
        sw = self.sample_weight_

        for j in xrange(self.n_clusters):
            mask = self.labels_ == j

            if np.sum(mask) == 0:
                raise ValueError("Empty cluster found, try smaller n_cluster.")

            denom = sw[mask].sum()
            denomsq = denom * denom

            if update_within:
                KK = K[mask][:, mask]  # K[mask, mask] does not work.
                dist_j = np.sum(np.outer(sw[mask], sw[mask]) * KK / denomsq)
                within_distances[j] = dist_j
                dist[:, j] += dist_j
            else:
                dist[:, j] += within_distances[j]

            dist[:, j] -= 2 * np.sum(sw[mask] * K[:, mask], axis=1) / denom

    def predict(self, X):
        K = self._get_kernel(X, self.X_fit_)
        n_samples = X.shape[0]
        dist = np.zeros((n_samples, self.n_clusters))
        self._compute_dist(K, dist, self.within_distances_,
                           update_within=False)
        return dist.argmin(axis=1)

if __name__ == '__main__':
    from sklearn.datasets import make_blobs
    X, y = make_blobs(n_samples=1000, centers=5, random_state=0)

    km = KernelKMeans(n_clusters=3, max_iter=100, random_state=0, verbose=1)
    print( km.fit_predict(X[:10]))
    print(km.predict(X[:10]))



"""
.. module:: Discretizer

Discretizer
*************

:Description: Discretizer

    

:Authors: bejar
    

:Version: 

:Created on: 13/03/2015 16:15 

"""

__author__ = 'bejar'

import numpy as np
from sklearn.base import TransformerMixin

#Todo: Add the possibility of using the (weighted) mean value of the interval
class Discretizer(TransformerMixin):
    """
    Discretization of the attributes of a dataset (unsupervised)

    method = 'equal' equal sized bins
             'frequency' bins with the same number of examples
    bins = number of bins
    """
    intervals = None

    def __init__(self, method='equal', bins=2):
        self.method = method
        self.bins = bins

    def _fit(self,X):
        """
        Computes the discretization intervals

        :param X:
        :return:
        """
        if self.method == 'equal':
            self._fit_equal(X)
        elif self.method == 'frequency':
            self._fit_frequency(X)

    def _fit_equal(self, X):
        """
        Computes the discretization intervals for equal sized discretization

        :param X:
        :return:
        """
        self.intervals = np.zeros((self.bins, X.shape[1]))

        for i in range(X.shape[1]):
            vmin = np.min(X[:, i])
            vmax = np.max(X[:, i])

            step = np.abs(vmax - vmin) / float(self.bins)
            for j in range(self.bins):
                vmin += step
                self.intervals[j, i] = vmin

            self.intervals[self.bins-1, i] += 0.00000000001

    def _fit_frequency(self, X):
        """
        Computes the discretization intervals for equal frequency

        :param X:
        :return:
        """

        self.intervals = np.zeros((self.bins, X.shape[1]))

        quant = X.shape[0] / float(self.bins)
        for i in range(X.shape[1]):
            lvals = sorted(X[:, i])
            nb = 0
            while nb < self.bins:
                self.intervals[nb, i] = lvals[int((quant*nb) + quant)-1]
                nb += 1
            self.intervals[self.bins-1, i] += 0.00000000001

    def _transform(self, X, copy=False):
        """
        Discretizes the attributes of a dataset
        :param X:
        :return:
        """
        if self.intervals is None:
            raise Exception('Discretizer: Not fitted')
        if copy:
            Y = X.copy()
        else:
            Y = X

        self.__transform(Y)

        return Y

    def __discretizer(self, v, at):
        """
        Determines the dicretized value for an atribute
        :param v:
        :return:
        """
        i=0
        while i< self.intervals.shape[0] and v > self.intervals[i, at]:
            i += 1
        return i

    def __transform(self, X):
        """
        Applies the discretization to all the attributes of the data matrix

        :param X:
        :return:
        """
        for i in range(X.shape[1]):
            for j in range(X.shape[0]):
                X[j, i] = self.__discretizer(X[j, i], i)

    def fit(self, X):
        """
        Fits a set of discretization intervals using the data in X
        :param X:
        :return:
        """
        self._fit(X)

    def transform(self, X, copy=False):
        """
        Applies previously fitted discretization intervals to X

        :param X:
        :param copy:
        :return:
        """

        return self._transform(X, copy=copy)


    def fit_transform(self, X, copy=False):
        """
        Fits and transforms the data

        :param X:
        :param copy:
        :return:
        """
        self._fit(X)
        return self._transform(X, copy=copy)




"""
.. module:: distances

distances
*************

:Description: distances

    Distances and divergence functions for probability matrices and vectors

:Authors: bejar
    

:Version: 

:Created on: 17/11/2014 13:01 

"""

__author__ = 'bejar'

import numpy as np


def simetrized_kullback_leibler_divergence(m1, m2):
    """
    Simetrized Kullback-Leibler divergence between two probability matrices/vectors
    :param m1:
    :param m2:
    :return:
    """
    lm1 = np.log(m1)
    lm2 = np.log(m2)
    lquot12 = np.log(m1 / m2)
    lquot21 = np.log(m2 / m1)
    dkl12 = lm1 * lquot12
    dkl21 = lm2 * lquot21
    return dkl12.sum() + dkl21.sum()

def kullback_leibler_divergence(m1, m2):
    """
    Kullback-Leibler divergence between two probability matrices/vectors
    :param m1:
    :param m2:
    :return:
    """
    lm1 = np.log(m1)
    lm2 = np.log(m2)
    lquot12 = np.log(m1 / m2)
    lquot21 = np.log(m2 / m1)
    dkl12 = lm1 * lquot12

    return dkl12.sum()

def jensen_shannon_divergence(m1, m2):
    """
    Jensen Shannon Divergence between two probability matrices/vectors

    :param m1:
    :param m2:
    :return:
    """
    m = 0.5*(m1+m2)

    return (0.5 * kullback_leibler_divergence(m1, m)) + (0.5 * kullback_leibler_divergence(m2, m))


def renyi_half_divergence(m1, m2):
    """
    Renyi divergence for parameter 1/2 between two probability matrices/vectors
    :param m1:
    :param m2:
    :return:
    """

    pm = m1 * m2
    spm = np.sqrt(pm)

    return -2 * np.log(spm.sum())


def square_frobenius_distance(m1, m2):
    """
    Square frobenius distance between two probability matrices/vectors
    :param m1:
    :param m2:
    :return:
    """
    c = m1 - m2
    c = c * c
    return c.sum()


def bhattacharyya_distance(m1, m2):
    """
    Bhattacharyya distance between two probability matrices/vectors
    :param m1:
    :param m2:
    :return:
    """
    sum = 0.0
    for a, b in zip(m1, m2):
        sum += np.sum(np.sqrt(a*b))
    return - np.log(sum)

def hellinger_distance(m1, m2):
    """
    Bhattacharyya distance between two probability matrices/vectors
    :param m1:
    :param m2:
    :return:
    """
    sum = 0.0
    for a, b in zip(m1, m2):
        sum += np.sum((np.sqrt(a) - np.sqrt(b)) ** 2)
    return (1/np.log(2)) * np.sqrt(sum)



"""
.. module:: Leader

Leader
*************

:Description: Leader Algorithm Clustering

    

:Authors: bejar
    

:Version: 

:Created on: 07/07/2014 8:29 

"""

__author__ = 'bejar'

import numpy as np
from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin
from sklearn.metrics.pairwise import euclidean_distances


class Leader(BaseEstimator, ClusterMixin, TransformerMixin):
    """Leader Algorithm Clustering

    Paramerets:

    radius: float
        Clustering radius for asigning examples to a cluster

    """

    def __init__(self, radius):
        self.radius = radius
        self.cluster_centers_ = None
        self.labels_ = None
        self.cluster_sizes_ = None

    def num_clusters(self):
        return self.cluster_centers_.shape[0]

    def fit(self, X):
        """
        Clusters the examples
        :param X:
        :return:
        """

        self.cluster_centers_, self.labels_, self.cluster_sizes_ = self._fit_process(X)

        return self

    def predict(self, X):
        """
        Returns the nearest cluster for a data matrix

        @param X:
        @return:
        """
        clasif = []
        for i in range(X.shape[0]):
            ncl, mdist = self._find_nearest_cluster(X[i].reshape(1, -1), self.cluster_centers_)
            if mdist <= self.radius:
                clasif.append(ncl)
            else:
                clasif.append(-1)
        return clasif

    def _fit_process(self, X):
        """
        Clusters incrementally the examples
        :param X:
        :return:
        """
        assignments = []
        scenters = np.zeros((1, X.shape[1]))
        centers = np.zeros((1, X.shape[1]))
        # Initialize with the first example
        scenters[0] = X[0]
        centers[0] = X[0]
        assignments.append([0])
        csizes = np.array([1])
        # Cluster the rest of examples
        for i in range(1, X.shape[0]):
            ncl, mdist = self._find_nearest_cluster(X[i].reshape(1, -1), centers)

            # if distance is less than radius, introduce example in nearest class
            if mdist <= self.radius:
                scenters[ncl] += X[i]
                csizes[ncl] += 1
                centers[ncl] = scenters[ncl] / csizes[ncl]
                assignments[ncl].append(i)
            else:  # Create a new cluster
                scenters = np.append(scenters, np.array([X[i]]), 0)
                centers = np.append(centers, np.array([X[i]]), 0)
                csizes = np.append(csizes, [1], 0)
                assignments.append([i])

        labels = np.zeros(X.shape[0])
        for l, ej in enumerate(assignments):
            for e in ej:
                labels[e] = l

        return centers, labels, csizes

    @staticmethod
    def _find_nearest_cluster(examp, centers):
        """
        Finds the nearest cluster for an example
        :param examp:
        :param centers:
        :return:
        """
        dist = euclidean_distances(centers, examp)

        pmin = np.argmin(dist)
        vmin = np.min(dist)

        return pmin, vmin

if __name__ == '__main__':
    from sklearn.datasets import make_blobs, load_iris, make_circles
    X, y_data = make_circles(n_samples=1000, noise=0.5, random_state=4, factor=0.5)
    ld = Leader(radius=.01)
    ld.fit(X)
    print ld.predict(np.array([[0,0]]))






"""
.. module:: test

test
*************

:Description: test

    

:Authors: bejar
    

:Version: 

:Created on: 25/11/2014 11:15 

"""

__author__ = 'bejar'

from LaplacianScore import LaplacianScore

from sklearn.datasets import make_blobs
import pylab as pl
import matplotlib.pyplot as plt


X,y_data = make_blobs(n_samples=5200, n_features=20, centers=7, random_state=0)

LS = LaplacianScore(n_neighbors=10, bandwidth=0.05)

LS.fit(X)

print(LS.scores_)
best = LS.best_k_scores(2)
print(best)

fig = plt.figure()

# ax = fig.gca(projection='3d')
# pl.scatter(X[:, 1], X[:, 2], zs=X[:, 0], c=ld.labels_, s=25)
#
ax = fig.add_subplot(111)
plt.scatter(X[:,best[0]],X[:,best[1]],c=y_data)

plt.show()




"""
.. module:: __init__.py

__init__.py
*************

:Description: __init__.py

    

:Authors: bejar
    

:Version: 

:Created on: 13/03/2015 16:13 

"""

__author__ = 'bejar'

from .Discretizer import Discretizer

__all__ = ['Discretizer']


"""
.. module:: SAX

SAX
*************

:Description: SAX

    

:Authors: bejar
    

:Version: 

:Created on: 18/03/2015 13:49 

"""

__author__ = 'bejar'


import numpy as np
from scipy.stats import norm

class SAX():
    """
    Sax representation of the time series
    """
    winlen = None
    step = None
    wrdlen = None
    voc = None
    intervals = None

    def __init__(self, window_length=100, step=1, word_length=10, voc_size=3):
        self.winlen = window_length
        self.step = step
        self.wrdlen = word_length
        self.voc = voc_size
        self.intervals = norm.ppf(np.arange(self.voc)/float(self.voc))


    # Todo: Check that this works
    def transform(self, X):
        """
        Computes the SAX representation for a vector of data
        The data is normalized before transformation

        Beware: If length is not a divisor of the vector size, some data
        points at the end will be ignored

        The intervals for the discretization are computed on every call
        :param length: Length of the wore
        :param voc: Length of the vocabulary
        :param intervals: list with the breakpoints of the discretization
                          if the parameter does not exist the intervals are
                          computed, but this is inefficient
        :return: a vector with length values in the range [-voc//2,voc//2]
        """

        res = np.zeros((X.shape[0]//self.step-self.winlen, self.wrdlen))

        for w in range(X.shape[0]//self.winlen):
            chunk = X[w*self.step, (w*self.step) + self.winlen]
            res[w] = self._SAX_function(chunk, self.wrdlen, self.voc, self.intervals)
        return res

    @staticmethod
    def _SAX_function(self, data, length, voc, intervals):
        """
        Computes the SAX representation for a vector of data
        The data is normalized before transformation

        Beware: If length is not a divisor of the vector size, some data
        points at the end will be ignored

        The intervals for the discretization are computed on every call
        :param length: Length of the wore
        :param voc: Length of the vocabulary
        :param intervals: list with the breakpoints of the discretization
                          if the parameter does not exist the intervals are
                          computed, but this is inefficient
        :return: a vector with length values in the range [-voc//2,voc//2]
        """
        index = np.zeros(length)
        data -= data.mean(0)
        data = np.nan_to_num(data / data.std(0))
        step = data.shape[0] / length
        for i in range(length):
            mr = np.sum(data[i*step:(i*step)+step])/length
            j = voc - 1
            while mr < intervals[j]:
                j -= 1
            index[i] = j - (voc / 2)
        return index



"""
.. module:: test

test
*************

:Description: test

    

:Authors: bejar
    

:Version: 

:Created on: 22/12/2014 14:34 

"""

__author__ = 'bejar'

from sklearn.datasets import load_iris, make_blobs
from quantization_error import quantization_error
from Xu import Xu
from cluster import jeffrey_divergence_score
from sklearn.cluster import KMeans
from kemlglearn.datasets import cluster_generator

nc = 5
X, y = cluster_generator(n_clusters=nc, sepval=0.2, numNonNoisy=20, numNoisy=3, rangeN=[150, 200])

#X, _ = make_blobs(n_samples=500, n_features=10, centers=5)

#X= load_iris()['data']

for nclust in range(2,10):
    km = KMeans(n_clusters=nclust)
    km.fit(X)
    v = jeffrey_divergence_score(X, km.labels_)
    print(nclust, v)

#qe.fit(X['data'])






"""
.. module:: __init__.py

__init__.py
*************

:Description: __init__.py

    

:Authors: bejar
    

:Version: 

:Created on: 22/12/2014 13:07 

"""

__author__ = 'bejar'

from .cluster import within_scatter_matrix_score, between_scatter_matrix_score, calinski_harabasz_score, zhao_chu_franti_score,\
    scatter_matrices_scores, davies_bouldin_score, variation_of_information_score, jaccard_score, \
    folkes_mallow_score, jeffrey_divergence_score

from .divergences import simetrized_kullback_leibler_divergence, kullback_leibler_divergence, bhattacharyya_distance,\
    jensen_shannon_divergence, renyi_half_divergence, square_frobenius_distance, hellinger_distance


__all__ = ['within_scatter_matrix_score',
           'between_scatter_matrix_score',
           'calinski_harabasz_score',
           'zhao_chu_franti_score',
           'scatter_matrices_scores',
           'davies_bouldin_score',
           'variation_of_information_score',
           'jaccard_score', 'JaccardArandFolkes',
           'folkes_mallow_score',
           'jeffrey_divergence_score',
           'simetrized_kullback_leibler_divergence',
           'kullback_leibler_divergence, bhattacharyya_distance',
           'jensen_shannon_divergence', 'renyi_half_divergence',
           'square_frobenius_distance', 'hellinger_distance']



"""
.. module:: GlobalKMeans

GlobalKMeans
*************

:Description: GlobalKMeans

    

:Authors: bejar
    

:Version: 

:Created on: 20/01/2015 10:42 

"""

__author__ = 'bejar'


import numpy as np
from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.cluster import KMeans
from sklearn.neighbors import NearestNeighbors


class GlobalKMeans(BaseEstimator, ClusterMixin, TransformerMixin):
    """Global K-means Algorithm

    Paramereters:

    n_clusters: int
        maximum number of clusters to obtain
    algorithm string
        'classical' the classical algorithm
        'bagirov' the Bagirov 2006 variant

    """
    def __init__(self, n_clusters, algorithm='classical'):
        self.n_clusters = n_clusters
        self.cluster_centers_ = None
        self.labels_ = None
        self.cluster_sizes_ = None
        self.inertia_ = None
        self.algorithm = algorithm

    def fit(self,X):
        """
        Clusters the examples
        :param X:
        :return:
        """

        if self.algorithm == 'classical':
            self.cluster_centers_, self.labels_, self.inertia_ = self._fit_process(X)
        elif self.algorithm == 'bagirov':
            self.cluster_centers_, self.labels_, self.inertia_ = self._fit_process_bagirov(X)

        return self

    def predict(self,X):
        """
        Returns the nearest cluster for a data matrix

        @param X:
        @return:
        """
        clasif = []
        for i in range(X.shape[0]):
            ncl, mdist = self._find_nearest_cluster(X[i], self.cluster_centers_)
            if mdist <= self.radius:
                clasif.append(ncl)
            else:
                clasif.append(-1)
        return clasif


    def _fit_process(self, X):
        """
        Classical global k-means algorithm

        :param X:
        :return:
        """

        # Compute the centroid of the dataset
        centroids = sum(X)/X.shape[0]
        centroids.shape = (1, X.shape[1])

        for i in range(2, self.n_clusters+1):
            mininertia = np.infty
            for j in range(X.shape[0]):
                newcentroids = np.vstack((centroids, X[j]))
                #print newcentroids.shape
                km = KMeans(n_clusters=i, init=newcentroids, n_init=1)
                km.fit(X)
                if mininertia > km.inertia_:
                    mininertia = km.inertia_
                    bestkm = km
            centroids = bestkm.cluster_centers_

        return bestkm.cluster_centers_, bestkm.labels_, bestkm.inertia_

    def _fit_process_bagirov(self, X):
        """
        Clusters using the global K-means algorithm Bagirov variation
        :param X:
        :return:
        """

        # Create a KNN structure for fast search
        self._neighbors = NearestNeighbors()
        self._neighbors.fit(X)

        # Compute the centroid of the dataset
        centroids = sum(X)/X.shape[0]
        assignments = [0 for i in range(X.shape[0])]

        centroids.shape = (1, X.shape[1])

        # compute the distance of the examples to the centroids
        mindist = np.zeros(X.shape[0])
        for i in range(X.shape[0]):
            mindist[i] = euclidean_distances(X[i], centroids[assignments[i]], squared=True)[0]

        for k in range(2, self.n_clusters+1):
            newCentroid = self._compute_next_centroid(X, centroids, assignments, mindist)
            centroids = np.vstack((centroids,newCentroid))
            km = KMeans(n_clusters=k, init=centroids, n_init=1)
            km.fit(X)
            assignments = km.labels_
            for i in range(X.shape[0]):
                mindist[i] = euclidean_distances(X[i], centroids[assignments[i]], squared=True)[0]

        return km.cluster_centers_, km.labels_, km.inertia_

    def _compute_next_centroid(self, X, centroids, assignments, mindist):
        """
        Computes the candidate for the next centroid

        :param X:
        :param centroids:
        :return:
        """
        minsum = np.infty
        candCentroid = None

        # Compute the first candidate to new centroid
        for i in range(X.shape[0]):
            distance = euclidean_distances(X[i], centroids[assignments[i]])[0]
            S2 = self._neighbors.radius_neighbors(X[i], radius=distance, return_distance=False)[0]
            S2centroid = np.sum(X[S2], axis=0)/len(S2)
            S2centroid.shape = (1, X.shape[1])
            cost = self._compute_fk(X, mindist, S2centroid)

            if cost < minsum:
                minsum = cost
                candCentroid = S2centroid

        # Compute examples for the new centroid
        S2 = []
        newDist = euclidean_distances(X, candCentroid, squared=True)
        for i in range(X.shape[0]):
            if newDist[i] < mindist[i]:
                S2.append(i)

        newCentroid = sum(X[S2])/len(S2)
        newCentroid.shape = (1, X.shape[1])

        while not (candCentroid == newCentroid).all():
            candCentroid = newCentroid
            S2 = []
            newDist = euclidean_distances(X, candCentroid, squared=True)
            for i in range(X.shape[0]):
                if newDist[i] < mindist[i]:
                    S2.append(i)

            newCentroid = np.sum(X[S2], axis=0)/len(S2)
            newCentroid.shape = (1, X.shape[1])

        return candCentroid


    def _compute_fk(self, X, mindist, ccentroid):
        """
        Computes the cost function

        :param X:
        :param mindist:
        :param ccentroid:
        :return:
        """

        # Distances among the examples and the candidate centroid
        centdist = euclidean_distances(X, ccentroid, squared=True)

        fk = 0
        for i in range(X.shape[0]):
            fk = fk + min(mindist[i], centdist[i][0])

        return fk


    @staticmethod
    def _find_nearest_cluster(examp, centers):
        """
        Finds the nearest cluster for an example
        :param examp:
        :param centers:
        :return:
        """

        dist = euclidean_distances(centers, examp)

        pmin = np.argmin(dist)
        vmin = np.min(dist)

        return pmin, vmin


"""
.. module:: test

test
*************

:Description: test


:Authors: bejar

:Version: 

:Created on: 10/02/2015 9:50 

"""

__author__ = 'bejar'

from MeanPartition import MeanPartitionClustering
from kemlglearn.datasets import cluster_generator
import matplotlib.pyplot as plt

from sklearn.datasets import make_blobs, load_iris, make_circles

nc = 15
X, y = cluster_generator(n_clusters=nc, sepval=0.2, numNonNoisy=100, numNoisy=10, rangeN=[150, 200])

# print X[:,:-2]
#
#print X,y


# X, y = make_blobs(n_samples=1000, n_features=20, centers=nc, random_state=2)


gkm = MeanPartitionClustering(n_clusters=nc, n_components=40, n_neighbors=3, trans='spectral', cdistance='ANMI')
res, l = gkm.fit(X, y)


fig = plt.figure()

# ax = fig.gca(projection='3d')
# pl.scatter(X[:, 1], X[:, 2], zs=X[:, 0], c=gkm.labels_, s=25)

ax = fig.add_subplot(111)
plt.scatter(res[:, 0], res[:, 1], c=l)

plt.show()




"""
.. module:: test

test
*************

:Description: test

    

:Authors: bejar
    

:Version: 

:Created on: 07/07/2014 11:12 

"""
__author__ = 'bejar'



from Leader import Leader
from GlobalKMeans import GlobalKMeans
from sklearn.datasets import make_blobs, load_iris, make_circles
import pylab as pl
import matplotlib.pyplot as plt
from sklearn.cluster import KMeans
from kemlglearn.datasets import cluster_generator
from kemlglearn.metrics import within_scatter_matrix_score, between_scatter_matrix_score, calinski_harabasz_score,\
    zhao_chu_franti_score, scatter_matrices_scores, davies_bouldin_score, variation_of_information_score, \
    jaccard_score, JaccardArandFolkes, folkes_mallow_score
from kemlglearn.cluster.consensus import SimpleConsensusClustering
from sklearn.metrics.cluster import normalized_mutual_info_score

#X, y_data = make_blobs(n_samples=100, n_features=10, centers=3, random_state=2)
#X = load_iris()['data']
X, y_data = make_circles(n_samples=1000, noise=0.5, random_state=4, factor=0.5)


# nc = 12
# _, X = cluster_generator(n_clusters=nc, sepval=0.01, numNonNoisy=15, numNoisy=3, rangeN=[50, 100])
# ld = Leader(radius=25.0)
#
# #print timeit.timeit(stmt='ld.fit(X)',setup=setup,number=10)
# ld.fit(X)
#
# # print ld.cluster_centers_.shape[0]
#
# fig = plt.figure()
#
# # ax = fig.gca(projection='3d')
# # pl.scatter(X[:, 1], X[:, 2], zs=X[:, 0], c=ld.labels_, s=25)
# #
# ax = fig.add_subplot(111)
# plt.scatter(X[:,0],X[:,1],c=ld.labels_)
#
# plt.show()

# gkm = GlobalKMeans(n_clusters=nc)
# gkm.fit(X)
# print DaviesBouldin(X, gkm.labels_)
# print scatter_matrices_scores(X, gkm.labels_, ['Inertia'])

#
# print gkm.cluster_centers_.shape[0], gkm.inertia_
# print within_scatter_matrix_score(X, gkm.labels_)
# print between_scatter_matrix_score(X, gkm.labels_)


# fig = plt.figure()
#
# # ax = fig.gca(projection='3d')
# # pl.scatter(X[:, 1], X[:, 2], zs=X[:, 0], c=gkm.labels_, s=25)
#
# ax = fig.add_subplot(111)
# plt.scatter(X[:, 0], X[:, 1], c=gkm.labels_)
#
# plt.show()
# for nc in range(2, 16):
#     km = KMeans(n_clusters=nc)
#     km.fit(X)
#     #print km.cluster_centers_.shape[0], km.inertia_
#     print nc, scatter_matrices_scores(X, km.labels_, ['CH', 'ZCF', 'Hartigan', 'Xu'])
#     print DaviesBouldin(X, km.labels_)

# fig = plt.figure()
#
# ax = fig.add_subplot(111)
# plt.scatter(X[:,0],X[:,1],c=km.labels_)
#
# plt.show()

km = KMeans(n_clusters=3)
km.fit(X)

# print variation_of_information_score(km.labels_, y_data)
# print normalized_mutual_info_score(km.labels_, y_data)
print(jaccard_score(km.labels_, y_data))
print(folkes_mallow_score(km.labels_, y_data))
print(JaccardArandFolkes(km.labels_, y_data))

# print DaviesBouldin(X, km.labels_)
# print scatter_matrices_scores(X, km.labels_, ['Inertia'])

# fig = plt.figure()
#
# ax = fig.add_subplot(111)
# plt.scatter(X[:,0],X[:,1],c=km.labels_)
#
# plt.show()


# simple = SimpleConsensusClustering(n_clusters=nc, n_components=40)
# simple.fit(X)
#
# print DaviesBouldin(X, simple.labels_)
# print scatter_matrices_scores(X, simple.labels_, ['Inertia'])

# fig = plt.figure()
#
# ax = fig.add_subplot(111)
# plt.scatter(X[:,0],X[:,1],c=simple.labels_)
#
# plt.show()


# simple = SimpleConsensusClustering(n_clusters=nc, n_components=40, consensus2='spectral')
# simple.fit(X)
#
# print DaviesBouldin(X, simple.labels_)
# print scatter_matrices_scores(X, simple.labels_, ['Inertia'])
#

# fig = plt.figure()
#
# ax = fig.add_subplot(111)
# plt.scatter(X[:, 0], X[:, 1], c=simple.labels_)
#
# plt.show()



"""
.. module:: __init__.py

__init__.py
*************

:Description: __init__.py

    

:Authors: bejar
    

:Version: 

:Created on: 21/01/2015 9:00 

"""

__author__ = 'bejar'

from .samples_generator import cluster_generator, make_blobs

__all__ = ['cluster_generator', 'make_blobs']



"""
.. module:: __init__.py

__init__.py
*************

:Description: __init__.py

:Authors: bejar

:Version:

:Created on: 07/07/2014 8:28 

"""

__author__ = 'bejar'

from .GlobalKMeans import GlobalKMeans
from .Leader import Leader
from .KernelKMeans import KernelKMeans
from .KModes import KModes
from .KPrototypes import KPrototypes

__all__ = ['GlobalKMeans',
           'Leader', 'KernelKMeans',
           'KModes', 'KPrototypes']
